#if !defined(AFX_SSS_H__CA82A541_5CD5_11DA_A928_444553540000__INCLUDED_)
#define AFX_SSS_H__CA82A541_5CD5_11DA_A928_444553540000__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif

class RLReadStream
{
protected:
	const ZWORD *ptr;
	ZWORD countv;

public:
	RLReadStream(const LIFEBUFFER &buffer)
	{
		ptr = buffer.data + 1;
		countv = *ptr++;
	}

	ZWORD count(void) const
	{
		return countv;
	}

	ZWORD next(void)
	{
		if (countv)
		{
			--countv;
			return 0;
		}
		else
		{
			ZWORD word = *ptr++;
			const ZWORD flags = word >> 30;

			if (flags)
			{
				countv = *ptr++;
				word &= 0x0FFFFFF0;
			}

			return word;
		}
	}

	void quickSkip(const ZWORD aCount)
	{
		countv -= aCount;
	}

	int atEnd(void) const
	{
		return !countv && !*ptr;
	}
};

class RLReadStreamCarryForward : public RLReadStream
{
private:
	ZWORD carry;

public:
	RLReadStreamCarryForward(const LIFEBUFFER &buffer)
		: RLReadStream(buffer)
	{
		carry = 0;
	}

	ZWORD count(void) const
	{
		return carry ? 0 : countv;
	}

	ZWORD next(void)
	{
		if (countv)
		{
			ZWORD temp = carry;
			carry = 0;
			--countv;
			return temp;
		}
		else
		{
			ZWORD word = *ptr++;
			const ZWORD flags = word >> 30;

			if (flags)
			{
				countv = *ptr++;
				word &= 0x0FFFFFF0;
			}

			{
				const ZWORD temp = carry;
				carry = word >> 24;
				return word | temp;
			}
		}
	}

	void quickSkip(const ZWORD aCount)
	{
		RLReadStream::quickSkip(aCount);
		if (aCount) carry = 0;
	}
};

class RLReadStreamCarryBackward : public RLReadStream
{
public:
	RLReadStreamCarryBackward(const LIFEBUFFER &buffer)
		: RLReadStream(buffer)
	{
	}

	ZWORD count(void) const
	{
		return countv && (*ptr & 0xF0) ? countv - 1 : countv;
	}

	ZWORD next(void)
	{
		if (countv)
		{
			if (!--countv) return (*ptr & 0xF0) << 20;
			else return 0;
		}
		else
		{
			ZWORD word = *ptr++;
			const ZWORD flags = word >> 30;

			if (flags)
			{
				countv = *ptr++;
				word &= 0x0FFFFFF0;
			}
			else word |= (*ptr & 0xF0) << 24;

			return word >> 4;
		}
	}
};

class RLReadStreamXShifted : public RLReadStream
{
private:
	ZWORD offset;
	ZWORD prev;

public:
	RLReadStreamXShifted(const LIFEBUFFER &buffer, const ZWORD xoffset)
		: RLReadStream(buffer)
	{
		offset = xoffset * 2;
		prev = 0;
	}

	ZWORD count(void) const
	{
		return prev ? 0 : countv;
	}

	ZWORD next(void)
	{
		if (countv)
		{
			ZWORD temp = prev;
			prev = 0;
			--countv;
			return temp;
		}
		else
		{
			ZWORD word = *ptr++;
			const ZWORD flags = word >> 30;

			if (flags)
			{
				countv = *ptr++;
				word &= 0x0FFFFFF0;
			}

			{
				const ZWORD temp = prev;
				prev = (word >> (24 - offset)) & 0x0FFFFFF0;
				return ((word << offset) & 0x0FFFFFF0) | temp;
			}
		}
	}

	void quickSkip(const ZWORD aCount)
	{
		RLReadStream::quickSkip(aCount);
		if (aCount) prev = 0;
	}
};

#endif // !defined(AFX_SSS_H__CA82A541_5CD5_11DA_A928_444553540000__INCLUDED_)
