#include "RLReadStream.h"

#define COUNT (*ptr)

//////////////////////////////////////////////////////////////////////

class RLWriteStream
{
private:
	ZWORD *ptr;

protected:
	LIFEBUFFER *bufferPtr;

public:
	RLWriteStream(LIFEBUFFER &buffer, const int jogged)
	{
		bufferPtr = &buffer;
		memset(&buffer, 0, sizeof(LIFEHEADER));
		buffer.jogged = jogged;
		ptr = buffer.data;
		*ptr++ = 0;
		COUNT = 0;
	}

	~RLWriteStream()
	{
		ptr[-1] |= 3 << 30;
		ptr++;
		*ptr++ = 0;
		bufferPtr->size = (char *) ptr - (char *) bufferPtr;
	}

	void displayDiff(const ZWORD newWord, const ZWORD oldWord) const
	{
	}

	void append(const ZWORD aWord)
	{
		if (!aWord) COUNT += 1;
		else
		{
			if (COUNT)
			{
				ptr[-1] |= 1 << 30;
				ptr++;
			}

			*ptr++ = aWord;
			COUNT = 0;
		}
	}

	void skip(const ZWORD aCount)
	{
		COUNT += aCount;
	}
};

//////////////////////////////////////////////////////////////////////

class RLWriteStreamCounting : public RLWriteStream
{
private:
	static ZBYTE Table[4096];

public:
	static void Initialize(void)
	{
		for (int i = 0; i < 4096; i++)
		{
			int c = 0;

			for (int s = 0; s < 12; s++)
				c += (i >> s) & 1;

			Table[i] = c;
		}
	}

	static int CountBits(ZWORD aWord)
	{
		return (int) Table[(aWord >> 4) & 0xFFF] +
			(int) Table[(aWord >> 16) & 0xFFF];
	}

public:
	RLWriteStreamCounting(LIFEBUFFER &buffer, const int jogged) :
		RLWriteStream(buffer, jogged)
	{
		buffer.flags |= POPULATIONMASK;
		buffer.population = 0;
	}

	void append(const ZWORD aWord)
	{
		if (aWord) bufferPtr->population += CountBits(aWord);

		RLWriteStream::append(aWord);
	}
};

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM> class RLWriteStreamOring : public WRITESTREAM
{
private:
	RLReadStream orIn;
	RLWriteStream orOut;

public:
	RLWriteStreamOring(LIFEBUFFER &buffer,
		LIFEBUFFER &orOutBuffer, LIFEBUFFER &orInBuffer) :
			WRITESTREAM(buffer, orInBuffer.jogged),
			orIn(orInBuffer), orOut(orOutBuffer, orInBuffer.jogged)
	{
	}

	void append(const ZWORD aWord)
	{
		orOut.append(orIn.next() | aWord);
		WRITESTREAM::append(aWord);
	}

	void skip(const ZWORD aCount)
	{
		ZWORD count = aCount;

		while (count)
		{
			const ZWORD n = orIn.count();

			if (n)
			{
				const ZWORD d = min(count, n);
				orIn.quickSkip(d);
				orOut.skip(d);
				count -= d;
			}
			else
			{
				orOut.append(orIn.next());
				count -= 1;
			}
		}

		WRITESTREAM::skip(aCount);
	}
};

typedef RLWriteStreamOring<RLWriteStreamCounting>
	RLWriteStreamCountingAndSmearing;

typedef RLWriteStreamOring<RLWriteStream>
	RLWriteStreamSmearing;

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM> class RLWriteDisplay :
	public WRITESTREAM, public RLStreamDisplay
{
public:
	RLWriteDisplay(LIFEBUFFER &buffer, const int jogged,
		const PDISPLAYPARAMS params) :
			WRITESTREAM(buffer, jogged),
			RLStreamDisplay(params, jogged)
	{
	}

	RLWriteDisplay(LIFEBUFFER &buffer,
		LIFEBUFFER &orOut, LIFEBUFFER &orIn, const PDISPLAYPARAMS params) :
			WRITESTREAM(buffer, orOut, orIn),
			RLStreamDisplay(params, orIn.jogged)
	{
	}

	void displayDiff(const ZWORD newWord, const ZWORD oldWord)
	{
		RLStreamDisplay::displayDiff(newWord, oldWord);
	}

	void append(const ZWORD aWord)
	{
		WRITESTREAM::append(aWord);
		RLStreamDisplay::skip(1);
	}

	void skip(const ZWORD aCount)
	{
		WRITESTREAM::skip(aCount);
		RLStreamDisplay::skip(aCount);
	}
};

//////////////////////////////////////////////////////////////////////

class RLWriteStreamCountingAndBounding : public RLWriteStreamCounting
{
public:
	static BOUNDINGBOXES BoxTable[4096];

public:
	static void Initialize(void)
	{
		for (int word = 1; word < 4096; word++)
		{
			BOUNDINGBOXES &entry = BoxTable[word];

			entry.ne = entry.e = entry.se = entry.s = -100;
			entry.sw = entry.w = entry.nw = entry.n = 100;
			
			for (int y = 0; y < 2; y++)
				for (int x = 0; x < 6; x++)
					if (word & (1 << (x + x + y)))
					{
						entry.ne = max(entry.ne, x - y + 1);
						entry.e = max(entry.e, x + 1);
						entry.se = max(entry.se, x + y + 1);
						entry.s = max(entry.s, y + 1);
						entry.sw = min(entry.sw, x - y);
						entry.w = min(entry.w, x);
						entry.nw = min(entry.nw, x + y);
						entry.n = min(entry.n, y);
					}
		}
	}

private:
	int offset;
	ZWORD position;

public:
	RLWriteStreamCountingAndBounding(LIFEBUFFER &buffer, const int jogged) :
		RLWriteStreamCounting(buffer, jogged)
	{
		BOUNDINGBOXES &boxes = bufferPtr->boxes;
		boxes.w = (int) (SIDE / 2);
		boxes.n = (int) (SIDE / 2);
		boxes.e = - (int) (SIDE / 2);
		boxes.s = - (int) (SIDE / 2);
		offset = (int) (SIDE / 2) + bufferPtr->jogged;
		position = 0;
	}

	void append(const ZWORD aWord)
	{
		if (aWord)
		{
			BOUNDINGBOXES &boxes = bufferPtr->boxes;
			const int x = (int) ((position % WIDTHINBRICKS) * 12) - offset;
			const int y = (int) ((position / WIDTHINBRICKS) * 2) - offset;

			if (ZWORD mask = (aWord & 0x0000FFF0) >> 4)
			{
				BOUNDINGBOXES &entry = BoxTable[mask];
				boxes.w = min(boxes.w, x + entry.w);
				boxes.n = min(boxes.n, y + entry.n);
				boxes.e = max(boxes.e, x + entry.e);
				boxes.s = max(boxes.s, y + entry.s);
			}

			if (ZWORD mask = (aWord & 0x0FFF0000) >> 16)
			{
				BOUNDINGBOXES &entry = BoxTable[mask];
				boxes.w = min(boxes.w, x + entry.w + 6);
				boxes.n = min(boxes.n, y + entry.n);
				boxes.e = max(boxes.e, x + entry.e + 6);
				boxes.s = max(boxes.s, y + entry.s);
			}
		}

		position++;

		RLWriteStreamCounting::append(aWord);
	}

	void skip(const ZWORD aCount)
	{
		RLWriteStream::skip(aCount);
		position += aCount;
	}
};

//////////////////////////////////////////////////////////////////////

#undef COUNT
