#include <stdlib.h>
#include <memory.h>
#include <minmax.h>

#include "RLife8.h"
#include "PlatformDisplay.h"
#include "RLReadStream.h"
#include "RLStreamDisplay.h"
#include "RLWriteStream.h"

#define TABLEBITS 16
#define LOOKUPBITS (TABLEBITS / 2)
#define TABLESIZE (1 << TABLEBITS)

ZBYTE GenTable[TABLESIZE];
ZWORD GenTopTable[TABLESIZE];
ZWORD GenBottomTable[TABLESIZE];

ZBYTE SmudgeTable[TABLESIZE];
ZWORD SmudgeTopTable[TABLESIZE];
ZWORD SmudgeBottomTable[TABLESIZE];

typedef struct
{
	const ZBYTE *main;
	const ZWORD *top;
	const ZWORD *bottom;
} TABLES;

TABLES GenTables = {GenTable, GenTopTable, GenBottomTable};
TABLES SmudgeTables = {SmudgeTable, SmudgeTopTable, SmudgeBottomTable};

ZWORD StaticBufferHolder1[16000];
ZWORD StaticBufferHolder2[16000];
ZWORD StaticBufferHolder3[16000];
ZWORD StaticBufferHolder4[16000];
ZWORD SmearBufferHolder1[16000];
ZWORD SmearBufferHolder2[16000];
ZWORD SmearBufferHolder3[16000];

LIFEBUFFER &StaticBuffer1 = (LIFEBUFFER &) *StaticBufferHolder1;
LIFEBUFFER &StaticBuffer2 = (LIFEBUFFER &) *StaticBufferHolder2;
LIFEBUFFER &StaticBuffer3 = (LIFEBUFFER &) *StaticBufferHolder3;
LIFEBUFFER &StaticBuffer4 = (LIFEBUFFER &) *StaticBufferHolder4;
LIFEBUFFER &SmearBuffer1 = (LIFEBUFFER &) *SmearBufferHolder1;
LIFEBUFFER &SmearBuffer2 = (LIFEBUFFER &) *SmearBufferHolder2;
LIFEBUFFER &SmearBuffer3 = (LIFEBUFFER &) *SmearBufferHolder3;

//////////////////////////////////////////////////////////////////////

void buildGenTable(void)
{
	static const int XLIMIT = LOOKUPBITS / 2 - 2;

	for (int i = 0; i < TABLESIZE; i++)
	{
		ZBYTE nr = 0;
		ZBYTE mr = 1;

		for (int xr = 1; xr <= XLIMIT; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;
				int c;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int xBit = x << 1;
						const int yBit = y < 2 ? y + LOOKUPBITS : y - 2;
						const int b = (i >> (xBit + yBit)) & 1;
						if (x == xr && y == yr) c = b;
						t += b;
					}

				if (t == 3 || (t == 4 && c)) nr |= mr;

				mr <<= 1;
			}

		GenTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

void buildGenTopTable(void)
{
	for (int i = 0; i < 65536; i++)
	{
		ZWORD nr = 0;
		ZWORD mr = 1;

		for (int xr = 1; xr <= 6; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;
				int c;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int b =
							y < 2 ? (i >> (x + x + y)) & 1 : 0;
						if (x == xr && y == yr) c = b;
						t += b;
					}

				if (t == 3 || (t == 4 && c)) nr |= mr;

				mr <<= 1;
			}

		GenTopTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

void buildGenBottomTable(void)
{
	for (int i = 0; i < 65536; i++)
	{
		ZWORD nr = 0;
		ZWORD mr = 1;

		for (int xr = 1; xr <= 6; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;
				int c;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int b =
							y < 2 ? 0 : (i >> (x + x + y - 2)) & 1;
						if (x == xr && y == yr) c = b;
						t += b;
					}

				if (t == 3 || (t == 4 && c)) nr |= mr;

				mr <<= 1;
			}

		GenBottomTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

void buildSmudgeTable(void)
{
	static const int XLIMIT = LOOKUPBITS / 2 - 2;

	for (int i = 0; i < TABLESIZE; i++)
	{
		ZBYTE nr = 0;
		ZBYTE mr = 1;

		for (int xr = 1; xr <= XLIMIT; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int xBit = x << 1;
						const int yBit = y < 2 ? y + LOOKUPBITS : y - 2;
						const int b = (i >> (xBit + yBit)) & 1;
						t += b;
					}

				if (t) nr |= mr;

				mr <<= 1;
			}

		SmudgeTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

void buildSmudgeTopTable(void)
{
	for (int i = 0; i < 65536; i++)
	{
		ZWORD nr = 0;
		ZWORD mr = 1;

		for (int xr = 1; xr <= 6; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int b =
							y < 2 ? (i >> (x + x + y)) & 1 : 0;
						t += b;
					}

				if (t) nr |= mr;

				mr <<= 1;
			}

		SmudgeTopTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

void buildSmudgeBottomTable(void)
{
	for (int i = 0; i < 65536; i++)
	{
		ZWORD nr = 0;
		ZWORD mr = 1;

		for (int xr = 1; xr <= 6; xr++)
			for (int yr = 1; yr <= 2; yr++)
			{
				int t = 0;

				for (int y = yr - 1; y <= yr + 1; y++)
					for (int x = xr - 1; x <= xr + 1; x++)
					{
						const int b =
							y < 2 ? 0 : (i >> (x + x + y - 2)) & 1;
						t += b;
					}

				if (t) nr |= mr;

				mr <<= 1;
			}

		SmudgeBottomTable[i] = nr;
	}
}

//////////////////////////////////////////////////////////////////////

int Empty(LIFEBUFFER &out)
{
	memset(&out, 0, sizeof(LIFEHEADER));

	out.period = 1;
	out.symmetryGroup = 0xFF;

	out.flags = HASHMASK | POPULATIONMASK | PERIODICMASK |
		BOUNDINGBOXMASK | NORMALIZERMASK | SYMMETRYGROUPMASK;

	out.data[0] = 3 << 30;
	out.data[1] = TOTALCOUNT;
	out.data[2] = 0;

	out.size = sizeof(LIFEHEADER) + 12;

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Copy(LIFEBUFFER &out, const LIFEBUFFER &in)
{
	memcpy(&out, &in, in.size);
	return 0;
}

//////////////////////////////////////////////////////////////////////

int SingleCell(LIFEBUFFER &out, const int x, const int y)
{
	memset(&out, 0, sizeof(LIFEHEADER));

	const int offset = SIDE >> 1;
	const ZWORD x0 = x + offset;
	const ZWORD y0 = y + offset;
	const ZWORD r = x0 % 12;
	const ZWORD c = (y0 >> 1) * (SIDE / 12) + x0 / 12;
	const ZWORD s = (r << 1) | (y0 & 1);
	const ZWORD m = 0x10 << s;

	out.data[0] = 2 << 30;
	out.data[1] = c;
	out.data[2] = m | (3 << 30);
	out.data[3] = TOTALCOUNT - c - 1;
	out.data[4] = 0;

	out.size = sizeof(LIFEHEADER) + 20;

	return 0;
}

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM, class READSTREAM>
void jogCommon(WRITESTREAM &result, READSTREAM &top, READSTREAM &bottom)
{
	while (!bottom.atEnd())
	{
		if (top.count())
		{
			if (bottom.count())
			{
				const ZWORD n = min(top.count(), bottom.count());

				top.quickSkip(n);
				bottom.quickSkip(n);
				result.skip(n);
			}
			else
			{
				top.quickSkip(1);
				result.append((bottom.next() << 3) & 0x0AAAAAA0);
			}
		}
		else
		{
			if (bottom.count())
			{
				bottom.quickSkip(1);
				result.append((top.next() << 1) & 0x05555550);
			}
			else
			{
				result.append(((top.next() << 1) & 0x05555550) |
					((bottom.next() << 3) & 0x0AAAAAA0));
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////

int Jog(LIFEBUFFER &out, const LIFEBUFFER &in)
{
	RLWriteStream result = RLWriteStream(out, !in.jogged);

	if (!in.jogged)
	{
		RLReadStreamCarryForward top = RLReadStreamCarryForward(in);
		RLReadStreamCarryForward bottom = RLReadStreamCarryForward(in);

		bottom.quickSkip(SIDE / 12);
		result.skip(SIDE / 12);

		jogCommon(result, top, bottom);
	}
	else
	{
		RLReadStreamCarryBackward top = RLReadStreamCarryBackward(in);
		RLReadStreamCarryBackward bottom = RLReadStreamCarryBackward(in);

		bottom.quickSkip(SIDE / 12);

		jogCommon(result, top, bottom);

		result.skip(SIDE / 12);
	}

	if (in.flags & POPULATIONMASK)
	{
		out.flags |= POPULATIONMASK;
		out.population = in.population;
	}

	if (in.flags & PERIODICMASK)
	{
		out.flags |= PERIODICMASK;
		out.period = in.period;
		out.xStride = in.xStride;
		out.yStride = in.yStride;
	}

	if (in.flags & BOUNDINGBOXMASK)
	{
		out.flags |= BOUNDINGBOXMASK;
		out.boxes.ne = in.boxes.ne;
		out.boxes.e = in.boxes.e;
		out.boxes.se = in.boxes.se;
		out.boxes.s = in.boxes.s;
		out.boxes.sw = in.boxes.sw;
		out.boxes.w = in.boxes.w;
		out.boxes.nw = in.boxes.nw;
		out.boxes.n = in.boxes.n;
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

const LIFEBUFFER &jogIf(int cond, const LIFEBUFFER &buffer,
	LIFEBUFFER &staticBuffer)
{
	if (cond)
	{
		Jog(staticBuffer, buffer);
		return staticBuffer;
	}
	else return buffer;
}

//////////////////////////////////////////////////////////////////////

int Compare(const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	const LIFEBUFFER &p1 = jogIf(in1.jogged, in1, StaticBuffer1);
	const LIFEBUFFER &p2 = jogIf(in2.jogged, in2, StaticBuffer2);

	return p1.size != p2.size
		? p1.size - p2.size
		: memcmp(p1.data, p2.data, p1.size - sizeof(LIFEHEADER));
}

//////////////////////////////////////////////////////////////////////

int Or(LIFEBUFFER &out, const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	RLReadStream s1 =
		RLReadStream(jogIf(in1.jogged > in2.jogged, in1, StaticBuffer1));
	RLReadStream s2 =
		RLReadStream(jogIf(in1.jogged < in2.jogged, in2, StaticBuffer1));
	RLWriteStream result = RLWriteStream(out, in1.jogged && in2.jogged);

	while (!s1.atEnd())
	{
		if (s1.count())
		{
			if (s2.count())
			{
				const ZWORD n = min(s1.count(), s2.count());

				s1.quickSkip(n);
				s2.quickSkip(n);
				result.skip(n);
			}
			else
			{
				s1.quickSkip(1);
				result.append(s2.next());
			}
		}
		else
		{
			if (s2.count())
			{
				s2.quickSkip(1);
				result.append(s1.next());
			}
			else
			{
				result.append(s1.next() | s2.next());
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Add(LIFEBUFFER &out, const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	RLReadStream s1 =
		RLReadStream(jogIf(in1.jogged > in2.jogged, in1, StaticBuffer1));
	RLReadStream s2 =
		RLReadStream(jogIf(in1.jogged < in2.jogged, in2, StaticBuffer1));
	RLWriteStream result = RLWriteStream(out, in1.jogged && in2.jogged);

	while (!s1.atEnd())
	{
		if (s1.count())
		{
			if (s2.count())
			{
				const ZWORD n = min(s1.count(), s2.count());

				s1.quickSkip(n);
				s2.quickSkip(n);
				result.skip(n);
			}
			else
			{
				s1.quickSkip(1);
				result.append(s2.next());
			}
		}
		else
		{
			if (s2.count())
			{
				s2.quickSkip(1);
				result.append(s1.next());
			}
			else
			{
				ZWORD w1 = s1.next();
				ZWORD w2 = s2.next();

				if (w1 & w2) return -1;

				result.append(w1 | w2);
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Xor(LIFEBUFFER &out, const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	RLReadStream s1 =
		RLReadStream(jogIf(in1.jogged > in2.jogged, in1, StaticBuffer1));
	RLReadStream s2 =
		RLReadStream(jogIf(in1.jogged < in2.jogged, in2, StaticBuffer1));
	RLWriteStream result = RLWriteStream(out, in1.jogged && in2.jogged);

	while (!s1.atEnd())
	{
		if (s1.count())
		{
			if (s2.count())
			{
				const ZWORD n = min(s1.count(), s2.count());

				s1.quickSkip(n);
				s2.quickSkip(n);
				result.skip(n);
			}
			else
			{
				s1.quickSkip(1);
				result.append(s2.next());
			}
		}
		else
		{
			if (s2.count())
			{
				s2.quickSkip(1);
				result.append(s1.next());
			}
			else
			{
				result.append(s1.next() ^ s2.next());
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Andnot(LIFEBUFFER &out, const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	RLReadStream s1 =
		RLReadStream(jogIf(in1.jogged > in2.jogged, in1, StaticBuffer1));
	RLReadStream s2 =
		RLReadStream(jogIf(in1.jogged < in2.jogged, in2, StaticBuffer1));
	RLWriteStream result = RLWriteStream(out, in1.jogged && in2.jogged);

	while (!s1.atEnd())
	{
		if (s1.count())
		{
			if (s2.count())
			{
				const ZWORD n = min(s1.count(), s2.count());

				s1.quickSkip(n);
				s2.quickSkip(n);
				result.skip(n);
			}
			else
			{
				s1.quickSkip(1);
				s2.next();
				result.append(0);
			}
		}
		else
		{
			if (s2.count())
			{
				s2.quickSkip(1);
				result.append(s1.next());
			}
			else
			{
				result.append(s1.next() & ~s2.next());
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM, class READSTREAM>
void next(WRITESTREAM &result, READSTREAM &top, READSTREAM &bottom,
	const TABLES &tables)
{
	static const ZWORD LOMASK = (1 << LOOKUPBITS) - 1;
	static const ZWORD HIMASK = LOMASK << LOOKUPBITS;

	while (!bottom.atEnd())
	{
		if (top.count())
		{
			if (bottom.count())
			{
				const ZWORD n = min(top.count(), bottom.count());

				top.quickSkip(n);
				bottom.quickSkip(n);
				result.skip(n);
			}
			else
			{
				top.quickSkip(1);

				ZWORD b = bottom.next();
				ZWORD n = 0;

				n |= tables.bottom[b & 0xFFFF] << 4;
				b = _rotr(b, 12);
				n |= tables.bottom[b & 0xFFFF] << 16;

				result.displayDiff(n, _rotr(b, 17) & 0x0AAAAAA0);
				result.append(n);
			}
		}
		else
		{
			if (bottom.count())
			{
				bottom.quickSkip(1);

				ZWORD a = top.next();
				ZWORD n = 0;

				n |= tables.top[a & 0xFFFF] << 4;
				a = _rotr(a, 12);
				n |= tables.top[a & 0xFFFF] << 16;

				result.displayDiff(n, _rotr(a, 19) & 0x05555550);
				result.append(n);
			}
			else
			{
				ZWORD a = _rotl(top.next(), LOOKUPBITS);
				ZWORD b = bottom.next();
				ZWORD n = 0;

				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 4;
				a = _rotr(a, 4);
				b = _rotr(b, 4);
				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 8;
				a = _rotr(a, 4);
				b = _rotr(b, 4);
				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 12;
				a = _rotr(a, 4);
				b = _rotr(b, 4);
				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 16;
				a = _rotr(a, 4);
				b = _rotr(b, 4);
				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 20;
				a = _rotr(a, 4);
				b = _rotr(b, 4);
				n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 24;

				result.displayDiff
					(n, (_rotr(a, LOOKUPBITS + 11) & 0x05555550) |
						(_rotr(b, 9) & 0x0AAAAAA0));
				result.append(n);
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM>
void next2(WRITESTREAM &result, const LIFEBUFFER &in,
	const TABLES &tables)
{
	if (!in.jogged)
	{
		RLReadStreamCarryForward top = RLReadStreamCarryForward(in);
		RLReadStreamCarryForward bottom = RLReadStreamCarryForward(in);

		bottom.quickSkip(SIDE / 12);
		result.skip(SIDE / 12);

		next(result, top, bottom, tables);
	}
	else
	{
		RLReadStreamCarryBackward top = RLReadStreamCarryBackward(in);
		RLReadStreamCarryBackward bottom = RLReadStreamCarryBackward(in);

		bottom.quickSkip(SIDE / 12);

		next(result, top, bottom, tables);

		result.skip(SIDE / 12);
	}
}

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM>
void genCommon(WRITESTREAM &result, const LIFEBUFFER &in)
{
	next2(result, in, GenTables);
}

//////////////////////////////////////////////////////////////////////

template<class WRITESTREAM>
void smudgeCommon(WRITESTREAM &result, const LIFEBUFFER &in)
{
	next2(result, in, SmudgeTables);
}

//////////////////////////////////////////////////////////////////////

void smudge(LIFEBUFFER &out, const LIFEBUFFER &in)
{
	RLWriteStream result = RLWriteStream(out, !in.jogged);
	smudgeCommon(result, in);
}

//////////////////////////////////////////////////////////////////////

int Smudge(LIFEBUFFER &out, const LIFEBUFFER &in, const int n)
{
	if (n <= 0) return -1;

	LIFEBUFFER *from;
	LIFEBUFFER *to = &(LIFEBUFFER &) in;

	for (int i = n; i--;)
	{
		from = to;
		to = i & 1 ? &StaticBuffer1 : &out;
		smudge(*to, *from);
	}

	return n;
}

//////////////////////////////////////////////////////////////////////

void gen(LIFEBUFFER &out, const LIFEBUFFER &in,
	const PDISPLAYPARAMS params)
{
	if (params)
	{
		RLWriteDisplay<RLWriteStream> result =
			RLWriteDisplay<RLWriteStream>(out, !in.jogged, params);
		genCommon(result, in);
	}
	else
	{
		RLWriteStream result = RLWriteStream(out, !in.jogged);
		genCommon(result, in);
	}
}

//////////////////////////////////////////////////////////////////////

void gen(LIFEBUFFER &out, const LIFEBUFFER &in,
	LIFEBUFFER &orOut, LIFEBUFFER &orIn,
	const PDISPLAYPARAMS params)
{
	if (params)
	{
		RLWriteDisplay<RLWriteStreamSmearing> result =
			RLWriteDisplay<RLWriteStreamSmearing>
				(out, orOut, orIn, params);
		genCommon(result, in);
	}
	else
	{
		RLWriteStreamSmearing result = RLWriteStreamSmearing(out, orOut, orIn);
		genCommon(result, in);
	}
}

//////////////////////////////////////////////////////////////////////

void countGen(LIFEBUFFER &out, const LIFEBUFFER &in,
	const PDISPLAYPARAMS params)
{
	if (params)
	{
		RLWriteDisplay<RLWriteStreamCounting> result =
			RLWriteDisplay<RLWriteStreamCounting>
				(out, !in.jogged, params);
		genCommon(result, in);
	}
	else
	{
		RLWriteStreamCounting result = RLWriteStreamCounting(out, !in.jogged);
		genCommon(result, in);
	}
}

//////////////////////////////////////////////////////////////////////

void countGen(LIFEBUFFER &out, const LIFEBUFFER &in,
	LIFEBUFFER &orOut, LIFEBUFFER &orIn,
	const PDISPLAYPARAMS params)
{
	if (params)
	{
		RLWriteDisplay<RLWriteStreamCountingAndSmearing> result =
			RLWriteDisplay<RLWriteStreamCountingAndSmearing>
				(out, orOut, orIn, params);
		genCommon(result, in);
	}
	else
	{
		RLWriteStreamCountingAndSmearing result =
			RLWriteStreamCountingAndSmearing(out, orOut, orIn);
		genCommon(result, in);
	}
}

//////////////////////////////////////////////////////////////////////

int quickGen(LIFEBUFFER &out, const LIFEBUFFER &in, LIFEBUFFER &smear,
	const int n, const PDISPLAYPARAMS params)
{
	if (n <= 0) return -1;

	LIFEBUFFER *from;
	LIFEBUFFER *to = &(LIFEBUFFER &) in;

	LIFEBUFFER *smearFrom = &SmearBuffer1;
	LIFEBUFFER *smearTo = &SmearBuffer2;
	LIFEBUFFER *smearBack = &SmearBuffer3;

	int i = n;

	Copy(*smearTo, in);
	Empty(*smearBack);
	smearBack->jogged = !in.jogged;

	do
	{
		from = to;
		to = --i & 1 ? &StaticBuffer1 : &out;

		{
			LIFEBUFFER *temp = smearBack;
			smearBack = smearTo;
			smearTo = smearFrom;
			smearFrom = temp;
		}

		gen(*to, *from, *smearTo, *smearFrom, params);

	} while (i && to->size != sizeof(LIFEHEADER) + 12);

	if (to != &out) Copy(out, *to);

	Or(smear, *smearTo, *smearBack);

	return n - i;
}
	
//////////////////////////////////////////////////////////////////////

int quickGen(LIFEBUFFER &out, const LIFEBUFFER &in, const int n,
	const PDISPLAYPARAMS params)
{
	if (n <= 0) return -1;

	LIFEBUFFER *from;
	LIFEBUFFER *to = &(LIFEBUFFER &) in;
	int i = n;

	do
	{
		from = to;
		to = --i & 1 ? &StaticBuffer1 : &out;
		gen(*to, *from, params);

	} while (i && to->size != sizeof(LIFEHEADER) + 12);

	if (to != &out) Copy(out, *to);

	return n - i;
}
	
//////////////////////////////////////////////////////////////////////

int tracingGen(LIFEBUFFER &out, const LIFEBUFFER &in, LIFEBUFFER &smear,
	const int n, const PDISPLAYPARAMS params, int &returnCode)
{
	if (n <= 0) return -1;

	LIFEBUFFER *from = 0;
	LIFEBUFFER *to = &(LIFEBUFFER &) in;
	LIFEBUFFER *prev;

	LIFEBUFFER *smearFrom = &SmearBuffer1;
	LIFEBUFFER *smearTo = &SmearBuffer2;
	LIFEBUFFER *smearBack = &SmearBuffer3;

	int pops[6];
	int index = 0;
	int count = 0;
	int i = n;

	Copy(*smearTo, in);
	Empty(*smearBack);
	smearBack->jogged = !in.jogged;

	memset(pops, 0, sizeof(pops));
	returnCode = 5;

	do
	{
		prev = from;
		from = to;
		i--;
		to =
			i % 3 == 0 ? &out :
			i % 3 == 1 ? &StaticBuffer3 :
			&StaticBuffer4;

		{
			LIFEBUFFER *temp = smearBack;
			smearBack = smearTo;
			smearTo = smearFrom;
			smearFrom = temp;
		}

		countGen(*to, *from, *smearTo, *smearFrom, params);

		const int pop = to->population;

		if (!pop) returnCode = 0;
		else if (pop != pops[index])
		{
			pops[index] = pop;
			count = 0;
		}
		else if (++count == 6)
		{
			returnCode =
				memcmp(pops, pops + 2, sizeof(pops[0]) * 4)
					?	3	:	// possible p3 pattern with/without ships
				Compare(*to, *prev)
					?	4	:	// possible p1/p2 pattern with ships
				pops[0] != pops[1] || Compare(*to, *from)
					?	2	:	// p2 pattern
						1	;	// stable pattern
		}

		if (++index == 6) index = 0;
	} while (i && returnCode == 5);

	if (to != &out) Copy(out, *to);

	Or(smear, *smearTo, *smearBack);

	if (returnCode >= 0 && returnCode <= 2)
	{
		out.flags |= PERIODICMASK;
		out.period = returnCode ? returnCode : 1;
		out.xStride = out.yStride = 0;
	}

	return n - i;
}

//////////////////////////////////////////////////////////////////////

int tracingGen(LIFEBUFFER &out, const LIFEBUFFER &in, const int n,
	const PDISPLAYPARAMS params, int &returnCode)
{
	if (n <= 0) return -1;

	LIFEBUFFER *from = 0;
	LIFEBUFFER *to = &(LIFEBUFFER &) in;
	LIFEBUFFER *prev;
	int pops[6];
	int index = 0;
	int count = 0;
	int i = n;

	memset(pops, 0, sizeof(pops));
	returnCode = 5;

	do
	{
		prev = from;
		from = to;
		i--;
		to =
			i % 3 == 0 ? &out :
			i % 3 == 1 ? &StaticBuffer3 :
			&StaticBuffer4;

		countGen(*to, *from, params);

		const int pop = to->population;

		if (!pop) returnCode = 0;
		else if (pop != pops[index])
		{
			pops[index] = pop;
			count = 0;
		}
		else if (++count == 6)
		{
			returnCode =
				memcmp(pops, pops + 2, sizeof(pops[0]) * 4)
					?	3	:	// possible p3 pattern with/without ships
				Compare(*to, *prev)
					?	4	:	// possible p1/p2 pattern with ships
				pops[0] != pops[1] || Compare(*to, *from)
					?	2	:	// p2 pattern
						1	;	// stable pattern
		}

		if (++index == 6) index = 0;
	} while (i && returnCode == 5);

	if (to != &out) memcpy(&out, to, to->size);

	if (returnCode >= 0 && returnCode <= 2)
	{
		out.flags |= PERIODICMASK;
		out.period = returnCode ? returnCode : 1;
		out.xStride = out.yStride = 0;
	}

	return n - i;
}

//////////////////////////////////////////////////////////////////////

int Gen(LIFEBUFFER &out, const LIFEBUFFER &in, LIFEBUFFER &smear, const int n,
	const PDISPLAYPARAMS params, int &returnCode)
{
	return &returnCode
		? &smear
			? tracingGen(out, in, smear, n, params, returnCode)
			: tracingGen(out, in, n, params, returnCode)
		: &smear
			? quickGen(out, in, smear, n, params)
			: quickGen(out, in, n, params);
}

//////////////////////////////////////////////////////////////////////

int FirstPoint(int out[], const LIFEBUFFER &in)
{
	const int offset = (int) (SIDE / 2) + in.jogged;
	RLReadStream stream = RLReadStream(in);

	const ZWORD count = stream.count();
	stream.quickSkip(count);

	if (!stream.atEnd())
	{
		const ZWORD word = stream.next();

		const int x0 = (int) (count % WIDTHINBRICKS * 12) - offset;
		const int y0 = (int) (count / WIDTHINBRICKS * 2) - offset;

		for (int y = 0; y < 2; y++)
			for (int x = 0; x < 12; x++)
				if (word & (0x10 << (x + x + y)))
				{
					out[0] = x0 + x;
					out[1] = y0 + y;
					return 0;
				}
	}

	return -1;
}

//////////////////////////////////////////////////////////////////////

int Points(int out[], LIFEBUFFER &in)
{
	const int offset = (int) (SIDE / 2) + in.jogged;
	RLReadStream stream = RLReadStream(in);
	ZWORD position = 0;
	int *ptr = out;

	while (true)
	{
		{
			const ZWORD count = stream.count();
			stream.quickSkip(count);
			position += count;
		}

		if (stream.atEnd()) break;

		while (const ZWORD word = (position++, stream.next()))
		{
			const ZWORD p = position - 1;
			const int x0 = (int) (p % WIDTHINBRICKS * 12) - offset;
			const int y0 = (int) (p / WIDTHINBRICKS * 2) - offset;

			for (int y = 0; y < 2; y++)
				for (int x = 0; x < 12; x++)
					if (word & (0x10 << (x + x + y)))
					{
						*ptr++ = x0 + x;
						*ptr++ = y0 + y;
					}
		}
	}

	in.flags |= POPULATIONMASK;
	in.population = (ptr - out) / 2;

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Display(const LIFEBUFFER &in, const PDISPLAYPARAMS params)
{
	const int offset = (int) (SIDE / 2) + in.jogged;
	RLReadStream stream = RLReadStream(in);
	ZWORD position = 0;

	PlatformDisplayOpen(params);

	while (true)
	{
		{
			const ZWORD count = stream.count();
			stream.quickSkip(count);
			position += count;
		}

		if (stream.atEnd()) break;

		while (const ZWORD word = (position++, stream.next()))
		{
			const ZWORD p = position - 1;
			const int x0 = (int) (p % WIDTHINBRICKS * 12) - offset;
			const int y0 = (int) (p / WIDTHINBRICKS * 2) - offset;

			for (int y = 0; y < 2; y++)
				for (int x = 0; x < 12; x++)
					if (word & (0x10 << (x + x + y)))
						PlatformDisplayPlot(x0 + x, y0 + y);
		}
	}

	PlatformDisplayClose();

	return 0;
}

//////////////////////////////////////////////////////////////////////

int BoundingBox(LIFEBUFFER &in)
{
	if (!(in.flags & BOUNDINGBOXMASK))
	{
		in.flags |= BOUNDINGBOXMASK;

		const int offset = (int) (SIDE / 2) + in.jogged;
		RLReadStream stream = RLReadStream(in);
		ZWORD position = 0;
		int ne = - (int) SIDE;
		int e = - (int) (SIDE / 2);
		int se = - (int) SIDE;
		int s = - (int) (SIDE / 2);
		int sw = (int) SIDE;
		int w = (int) (SIDE / 2);
		int nw = (int) SIDE;
		int n = (int) (SIDE / 2);
		int empty = 1;

		while (true)
		{
			{
				const ZWORD count = stream.count();
				stream.quickSkip(count);
				position += count;
			}

			if (stream.atEnd()) break;

			while (const ZWORD word = (position++, stream.next()))
			{
				const ZWORD p = position - 1;
				const int x = (int) ((p % WIDTHINBRICKS) * 12) - offset;
				const int y = (int) ((p / WIDTHINBRICKS) * 2) - offset;

				if (ZWORD mask = (word & 0x0000FFF0) >> 4)
				{
					BOUNDINGBOXES &entry = RLWriteStreamCountingAndBounding::BoxTable[mask];
					ne = max(ne, x - y + entry.ne);
					e = max(e, x + entry.e);
					se = max(se, x + y + entry.se);
					s = max(s, y + entry.s);
					sw = min(sw, x - y + entry.sw);
					w = min(w, x + entry.w);
					nw = min(nw, x + y + entry.nw);
					n = min(n, y + entry.n);
				}

				if (ZWORD mask = (word & 0x0FFF0000) >> 16)
				{
					BOUNDINGBOXES &entry = RLWriteStreamCountingAndBounding::BoxTable[mask];
					int x6 = x + 6;
					ne = max(ne, x6 - y + entry.ne);
					e = max(e, x6 + entry.e);
					se = max(se, x6 + y + entry.se);
					s = max(s, y + entry.s);
					sw = min(sw, x6 - y + entry.sw);
					w = min(w, x6 + entry.w);
					nw = min(nw, x6 + y + entry.nw);
					n = min(n, y + entry.n);
				}

				empty = 0;
			}
		}

		if (empty) ne = e = se = s = sw = w = nw = n = 0;

		in.boxes.ne = ne;
		in.boxes.e = e;
		in.boxes.se = se;
		in.boxes.s = s;
		in.boxes.sw = sw;
		in.boxes.w = w;
		in.boxes.nw = nw;
		in.boxes.n = n;
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int CornerPost(LIFEBUFFER &in)
{
	const int offset = (int) (SIDE / 2) + in.jogged;
	RLReadStream stream = RLReadStream(in);
	ZWORD position = 0;
	int post = - (int) SIDE * 2;
	int empty = 1;

	while (true)
	{
		{
			const ZWORD count = stream.count();
			stream.quickSkip(count);
			position += count;
		}

		if (stream.atEnd()) break;

		while (const ZWORD word = (position++, stream.next()))
		{
			const ZWORD p = position - 1;
			const int x0 = (int) ((p % WIDTHINBRICKS) * 12) - offset;
			const int y0 = (int) ((p / WIDTHINBRICKS) * 2) - offset;

			for (int y = 0; y < 2; y++)
				for (int x = 0; x < 12; x++)
					if (word & (0x10 << (x + x + y)))
						post = max(post, 2 * (x0 + x + y0 + y + 2));

			empty = 0;
		}
	}

	if (empty) post = 0;

	return post;
}

//////////////////////////////////////////////////////////////////////

int Population(LIFEBUFFER &in)
{
	if (!(in.flags & POPULATIONMASK))
	{
		in.flags |= POPULATIONMASK;

		RLReadStream stream = RLReadStream(in);
		int population = 0;

		while (true)
		{
			stream.quickSkip(stream.count());

			if (stream.atEnd()) break;

			while (const ZWORD word = stream.next())
				population += RLWriteStreamCounting::CountBits(word);
		}

		in.population = population;
	}

	return in.population;
}

//////////////////////////////////////////////////////////////////////

int And(LIFEBUFFER &out, const LIFEBUFFER &in1, const LIFEBUFFER &in2)
{
	RLReadStream s1 =
		RLReadStream(jogIf(in1.jogged > in2.jogged, in1, StaticBuffer1));
	RLReadStream s2 =
		RLReadStream(jogIf(in1.jogged < in2.jogged, in2, StaticBuffer1));
	RLWriteStream result = RLWriteStream(out, in1.jogged && in2.jogged);

	while (!s1.atEnd())
	{
		if (s1.count())
		{
			if (s2.count())
			{
				const ZWORD n = min(s1.count(), s2.count());

				s1.quickSkip(n);
				s2.quickSkip(n);
				result.skip(n);
			}
			else
			{
				s1.quickSkip(1);
				s2.next();
				result.append(0);
			}
		}
		else
		{
			if (s2.count())
			{
				s1.next();
				s2.quickSkip(1);
				result.append(0);
			}
			else
			{
				result.append(s1.next() & s2.next());
			}
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Translate(LIFEBUFFER &out, const LIFEBUFFER &in,
	const int x, const int y)
{
	const int adjust = !(y & 1) ? 0 : in.jogged ? -1 : 1;

	const ZWORD xpos = (ZWORD) (x + adjust) + SIDE / 2;

	RLReadStreamXShifted s = RLReadStreamXShifted(in, xpos % 12);
	RLWriteStream result =
		RLWriteStream(out, y & 1 ? !in.jogged : in.jogged);

	const ZWORD count = s.count();
	s.quickSkip(count);

	if (s.atEnd())
	{
		result.skip(count);
	}
	else
	{
		const ZWORD xoffset = (ZWORD) (xpos / 12) - WIDTHINBRICKS / 2;
		const ZWORD yoffset = (ZWORD) (y + adjust) * (WIDTHINBRICKS / 2);
		const ZWORD offset = xoffset + yoffset;

		result.skip(count + offset);

		while (true)
		{
			{
				const ZWORD count = s.count();
				s.quickSkip(count);
				result.skip(count);
			}

			if (s.atEnd()) break;

			while (true)
			{
				const ZWORD word = s.next();
				result.append(word);
				if (!word) break;
			}
		}

		result.skip(0 - offset);
	}

	if (in.flags & POPULATIONMASK)
	{
		out.flags |= POPULATIONMASK;
		out.population = in.population;
	}

	if (in.flags & PERIODICMASK)
	{
		out.flags |= PERIODICMASK;
		out.period = in.period;
		out.xStride = in.xStride;
		out.yStride = in.yStride;
	}

	if (in.flags & BOUNDINGBOXMASK)
	{
		out.flags |= BOUNDINGBOXMASK;
		out.boxes.ne = in.boxes.ne + x - y;
		out.boxes.e = in.boxes.e + x;
		out.boxes.se = in.boxes.se + x + y;
		out.boxes.s = in.boxes.s + y;
		out.boxes.sw = in.boxes.sw + x - y;
		out.boxes.w = in.boxes.w + x;
		out.boxes.nw = in.boxes.nw + x + y;
		out.boxes.n = in.boxes.n + y;
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////

int Hash(LIFEBUFFER &in)
{
	if (!(in.flags & HASHMASK))
	{
		in.flags |= HASHMASK;

		const LIFEBUFFER &buffer = jogIf(in.jogged, in, StaticBuffer1);

		ZWORD hash = 0;

		for (ZWORD i = 1; i < (buffer.size - sizeof(LIFEHEADER)) / 4; i++)
		{
			ZWORD k = buffer.data[i];
			hash = 5 * hash + k + (k << 13);
		}

		in.hash = hash & 0x1FFFFFFF;
	}

	return in.hash;
}

//////////////////////////////////////////////////////////////////////

void Initialize(void)
{
	buildGenTable();
	buildGenTopTable();
	buildGenBottomTable();

	buildSmudgeTable();
	buildSmudgeTopTable();
	buildSmudgeBottomTable();

	RLWriteStreamCounting::Initialize();
	RLWriteStreamCountingAndBounding::Initialize();
}
